// number.e
//
// Fri, 02/26/1993 16:53:13
//
// Purpose:
//
// 		This extension inserts ordinal numbers one by one in the
// current buffer between point and mark (which may be 0 and size for
// the number_buffer command).
//
// Usage:
//
// 		Load this module and save your state file.  The command will
// prompt for the insertion column, the starting value and the number
// format.  The command then queries the user for each inserted
// number.
//
//	Y or <Space> inserts and goes to the next insert.
//	N or <Backspace> doesn't insert, but goes to the next insert.
//	<Esc> exits immediately.
//	! inserts into all remaining lines.
//	# prompts for a new number, inserts and continues
//	C-R enters a recursive edit, allowing you to modify the buffer
//		arbitrarily.  When you exit the recursive edit with
//		exit-level, the insertions continue.
//
// The user varable number_format is used to determine the output
// format of the inserted number.  This format is in the standard
// printf format.  Several different format strings may be useful:
//
//	"%d"		will use the number of characters required for number
// 	"%-10d"		will place numbers left justified in a field of
// 				length 10 using a decimal notation.
// 	"%08x"		will print the number as a HEX value in a field of
// 				width 8 and will fill the extra with zeros.
// 	"%02.8dL"	will print the numbers in a minimum field of width 2
// 				and a maximum field of width 8.  If the number width
// 				is less than the minimun then zeros will be appended.
// 				Finally, an L will be appended to the number.
//
//
//
// The modifications to this code are dedicated to the public domain
// with the caveat that Lugaru is welcome to use this within their
// distribution source code which is supplied with Epsilon.  Use it
// any way you want for whatever purposes.
//
//
//	John Kercheval
//	15563 10th Ave NE
//	Seattle, WA 98155-6209
//	INTERNET: johnk@wrq.com
//	Compu$erve: 72450,3702
//
// Modification History:
//
//		Fri, 02/26/1993 15:50:21 V1.0 initial release.
//		Thu, 07/08/1993 09:23:04 V1.1 prevent problems if the endpoint
//									the end of the file.
//		Tue, 07/20/1993 18:27:00 V1.2 fix no abort/highlight bug - AAC
//									Aaron A. Collins @ CIS 70324,3200
//		Thu, 07/22/1993 22:22:28 V1.3 fix minor bugs - Lugaru
//		Fri, 09/16/1994 11:20:11 V1.4 update for V7.0 of Epsilon, modify
//									numberFormat variable to look like other
//									Epsilon variables
//		Sat, 12/31/1994 11:02:58 V1.5 corrected problem with '?'
//

#include <eel.h>


#define BOOLEAN int
#define TRUE 1
#define FALSE 0

#define _INSERTED		0
#define _SKIPPED		1
#define _CONTINUOUS		2
#define _ABORT			3

//
// this variable is used to determine the format which is used to
// insert the number into the buffer.  This format is in the standard
// printf format.  Several different formats may be useful:
//
//	"%d"		will use the number of characters required for number
//	"%-10d"		will place numbers left justified in a field of
//				length 10 using a decimal notation.
//	"%08x"		will print the number as a HEX value in a field of
//				width 8 and will fill the extra with zeros.
//	"%02.8dL"	will print the numbers in a minimum field of width 2
//				and a maximum field of width 8.  If the number width
//				is less than the minimun then zeros will be appended.
//				Finally, an L will be appended to the number.
//
user char number_format[80] = "%d";


////////////////////////////////////////////////////////////////////////////
//
// number_buffer() will prompt for a column, a starting number and
// will number all lines in the file (via a query replace type of
// interface).  This will always insert numbers in whitespace at or
// after the given column.
//
//
command number_buffer()
{
	//
	// do the numbering for the whole buffer
	//
	do_number_prompt(0, size());
}


////////////////////////////////////////////////////////////////////////////
//
// number_region() will prompt for a column, a starting number and
// will number all lines in the region (via a query replace type of
// interface).  This will always insert numbers in whitespace at or
// after the given column.
//
//
command number_region()
{
	//
	// do the numbering for the current region
	//
	do_number_prompt(point, mark);
}


////////////////////////////////////////////////////////////////////////////
//
// do_number_prompt() is a helper function for the number and
// number_region command.
//
//
void do_number_prompt(start, end)
int start;						 // start of region
int end;						 // end of region
{
	char prompt[80];			 // query prompt
	int defaultColumn;			 // column number default
	int numColumn;				 // column number to use
	int firstNumber;			 // number to start at

	save_spot point, mark;

	//
	// set the default
	//
	if (has_arg) {
		defaultColumn = iter;
		iter = 0;
	} else {
		defaultColumn = current_column();
	}
	//
	// prompt for the column number
	//
	sprintf(prompt, "Column number [%d]: ", defaultColumn);
	numColumn = get_number(prompt);
	if (got_bad_number) {
		numColumn = defaultColumn;
	}
	//
	// prompt for the starting number
	//
	firstNumber = get_number("Base number [1]: ");
	if (got_bad_number) {
		firstNumber = 1;
	}
	//
	// prompt for the number format
	//
	strcpy(prompt, number_format);
	get_strdef(number_format, "Number format", prompt);

	//
	// perform the numbering
	//
	do_number_region(start, end, numColumn, firstNumber);
}


////////////////////////////////////////////////////////////////////////////
//
// do_number_region() will insert ordinal numbers from start to end
// in the given column and starting at base.  This will be done via a
// subset of the query_replace interface (ie ! will do all the rest
// of the region, Y will do 1, N will skip this line and you may
// directly specify the next number).
//
void do_number_region(start, end, column, base)
int start;						 // start of region
int end;						 // end of region
int column;						 // default column
int base;						 // starting value
{
	spot startSpot, endSpot;
	BOOLEAN query = TRUE;
	int numAdded = 0;
	int rval;

	//
	// validate the proper relationship of start and end
	//
	if (start > end) {
		rval = start;
		start = end;
		end = rval;
	}

	//
	// allocate the spots
	//
	point = start;
	startSpot = alloc_spot();
	point = end;
	endSpot = alloc_spot();

	//
	// place the numbers
	//
	for (point = *startSpot;		  point < *endSpot; nl_forward()) {
        if (current_column() <= column) {     /* AAC  ^ this was a comma... */
			force_to_column(column);

			rval = number_insert(&base, query);
			switch (rval) {
				case _CONTINUOUS:
					query = FALSE;
					//
					// fall through
					//
				case _INSERTED:
					base++;
					numAdded++;
					break;

				case _SKIPPED:
					break;

				case _ABORT:
					point = *endSpot;
					say("Number inserts aborted.");
					break;

				default:
					break;

			}
		}
	}

	//
	// show results
	//
	if (numAdded != 1) {
		say("%d lines modified.", numAdded);
	} else {
		say("%d line modified.", numAdded);
	}

	//
	// free the spots
	//
	free_spot(startSpot);
	free_spot(endSpot);
}


////////////////////////////////////////////////////////////////////////////
//
// number_insert() will insert number at the current location using
// the query_replace (a limited subset) type of interface.
//
// Y or <Space> inserts and goes to the next insert.
// N or <Backspace> doesn't insert, but goes to the next insert.
// <Esc> exits immediately.
// ! inserts into all remaining lines.
// # prompts for a new number, inserts and continues
// C-R enters a recursive edit, allowing you to modify the buffer
//     arbitrarily.  When you exit the recursive edit with
//     exit-level, the insertions continue.
//
//
int number_insert(number, query)
	int *number;
{
	//
	// prompt the user if query is true
	//
	if (query) {
		//
		// this label is jumped to if some interrupting action occurs
		//
	  ask:

		//
		// place the update notice and highlight the line
		//
		say("Insert \"%d\"", *number);
		make_line_highlight();
		maybe_refresh();		/* show the insertion location */
		//
		// obtain the key, resolve it if it is ctrl prefixed and
		// remove the highlight
		//
		getkey();
		if (key == CTRL('^')) {
			full_getkey("", CTRLIFY_KEY);
		}
		remove_line_highlight();
		modify_region(1, MRCOLOR, color_class highlight); // restore color
		//
		// is this a help key specified
		//
		if (key == '?' || key == 'h' || root_keys[key] == (short) help) {
			help_on_command("number-region");
			goto ask;
		}
		//
		// check for user abort
		//
		if (user_abort || key == abort_key) {
			return _ABORT;
		}
		//
		// check for standard key strokes
		//
		switch(tolower(key)) {
			case '\b':
			case 'n':
				//
				// do not insert this line
				//
				return _SKIPPED;
				break;

			case '!':
				//
				// do not query further
				//
				bprintf(number_format, *number);
				return _CONTINUOUS;
				break;

			case '#': {
				char prompt[80];			 // query prompt
				int nextNumber = *number;
				//
				// obtain a new base number
				//
				sprintf(prompt, "New base number [%d]: ", nextNumber);
				*number = get_number(prompt);
				if (got_bad_number) {
					*number = nextNumber;
				}
				//
				// fall through
				//
				case ' ':
				case 'y':
					//
					// do the insert
					//
					bprintf(number_format, *number);
				return _INSERTED;
				break;

				case CTRL('R'): {
					save_spot point, mark;

					//
					// recursive edit
					//
					say("");
					recursive_edit();
					goto ask;
					break;
				}

				case ESC:
				default:
					return _ABORT;
				break;
			}
		}
	} else {
		bprintf(number_format, *number);
		return _INSERTED;
	}
}
